from .types import (
    AgentMessage,
    AgentTool,
    AgentToolResult,
    AgentContext,
    AgentState,
    AgentLoopConfig,
    AgentEvent,
    AgentStartEvent,
    AgentEndEvent,
    TurnStartEvent,
    TurnEndEvent,
    MessageStartEvent,
    MessageUpdateEvent,
    MessageEndEvent,
    ToolExecutionStartEvent,
    ToolExecutionUpdateEvent,
    ToolExecutionEndEvent,
    ThinkingLevel,
    StreamFn
)
from .agent_loop import agent_loop, agent_loop_continue
from .agent import Agent, AgentOptions

__all__ = [
    "AgentMessage",
    "AgentTool",
    "AgentToolResult",
    "AgentContext",
    "AgentState",
    "AgentLoopConfig",
    "AgentEvent",
    "AgentStartEvent",
    "AgentEndEvent",
    "TurnStartEvent",
    "TurnEndEvent",
    "MessageStartEvent",
    "MessageUpdateEvent",
    "MessageEndEvent",
    "ToolExecutionStartEvent",
    "ToolExecutionUpdateEvent",
    "ToolExecutionEndEvent",
    "ThinkingLevel",
    "StreamFn",
    "agent_loop",
    "agent_loop_continue",
    "Agent",
    "AgentOptions"
]



"""
Stateful Agent class wrapper.
Manages the agent's internal state, tool bindings, event subscriptions, and queues for steering and follow-ups.
Provides `prompt()`, `continue_session()`, `steer()`, and `follow_up()` public accessors.
"""

import asyncio
from typing import List, Optional, Callable, Set, Union, Literal, Awaitable, Dict
import time

from sentarc_ai.models import get_model
from sentarc_ai.stream import stream_simple
from sentarc_ai.types import (
    Message,
    ModelDef,
    ImageContent,
    ReasoningEffort
)

from sentarc_agent.types import (
    AgentState,
    AgentOptions,
    AgentMessage,
    AgentEvent,
    AgentTool,
    AgentContext,
    AgentLoopConfig,
    ThinkingLevel,
    StreamFn
)
from sentarc_agent.agent_loop import agent_loop, agent_loop_continue

def default_convert_to_llm(messages: List[AgentMessage]) -> List[Message]:
    """Default converter: Keep only LLM-compatible messages."""
    return [
        m for m in messages if isinstance(m, dict) and m.get("role") in ["user", "assistant", "toolResult"]
        or getattr(m, "role", None) in ["user", "assistant", "toolResult"]
    ] # type: ignore


class Agent:
    def __init__(self, options: Optional[AgentOptions] = None):
        options = options or AgentOptions()
        
        self._state = AgentState(
            system_prompt="",
            model=get_model("google", "gemini-2.5-flash-lite-preview-06-17"),
            thinking_level="off",
            tools=[],
            messages=[],
            is_streaming=False,
            stream_message=None,
            pending_tool_calls=set(),
            error=None
        )
        if options.initial_state:
            for k, v in options.initial_state.items():
                if hasattr(self._state, k):
                    setattr(self._state, k, v)

        self.convert_to_llm = options.convert_to_llm or default_convert_to_llm
        self.transform_context = options.transform_context
        self.steering_mode = options.steering_mode or "one-at-a-time"
        self.follow_up_mode = options.follow_up_mode or "one-at-a-time"
        self.stream_fn = options.stream_fn or stream_simple
        self._session_id = options.session_id
        self._thinking_budgets = options.thinking_budgets
        self.get_api_key = options.get_api_key
        
        self.listeners: Set[Callable[[AgentEvent], None]] = set()
        self.abort_controller: Optional[asyncio.Event] = None
        
        self.steering_queue: List[AgentMessage] = []
        self.follow_up_queue: List[AgentMessage] = []
        self.running_prompt: Optional[asyncio.Task] = None

    @property
    def session_id(self) -> Optional[str]:
        return self._session_id

    @session_id.setter
    def session_id(self, value: Optional[str]):
        self._session_id = value

    @property
    def thinking_budgets(self) -> Optional[Dict[str, int]]:
        return self._thinking_budgets

    @thinking_budgets.setter
    def thinking_budgets(self, value: Optional[Dict[str, int]]):
        self._thinking_budgets = value

    @property
    def state(self) -> AgentState:
        return self._state

    def subscribe(self, fn: Callable[[AgentEvent], None]) -> Callable[[], None]:
        self.listeners.add(fn)
        def unsubscribe():
            self.listeners.discard(fn)
        return unsubscribe

    def set_system_prompt(self, v: str):
        self._state.system_prompt = v

    def set_model(self, m: ModelDef):
        self._state.model = m

    def set_thinking_level(self, l: ThinkingLevel):
        self._state.thinking_level = l

    def set_steering_mode(self, mode: Literal["all", "one-at-a-time"]):
        self.steering_mode = mode

    def set_follow_up_mode(self, mode: Literal["all", "one-at-a-time"]):
        self.follow_up_mode = mode

    def set_tools(self, t: List[AgentTool]):
        self._state.tools = t

    def replace_messages(self, ms: List[AgentMessage]):
        self._state.messages = list(ms)

    def append_message(self, m: AgentMessage):
        self._state.messages.append(m)

    def steer(self, m: AgentMessage):
        """Queue a steering message to interrupt the agent mid-run."""
        self.steering_queue.append(m)

    def follow_up(self, m: AgentMessage):
        """Queue a follow-up message to be processed after the agent finishes."""
        self.follow_up_queue.append(m)

    def clear_steering_queue(self):
        self.steering_queue.clear()

    def clear_follow_up_queue(self):
        self.follow_up_queue.clear()

    def clear_all_queues(self):
        self.steering_queue.clear()
        self.follow_up_queue.clear()

    def has_queued_messages(self) -> bool:
        return len(self.steering_queue) > 0 or len(self.follow_up_queue) > 0

    def _dequeue_steering_messages(self) -> List[AgentMessage]:
        if self.steering_mode == "one-at-a-time":
            if self.steering_queue:
                return [self.steering_queue.pop(0)]
            return []
        steering = list(self.steering_queue)
        self.steering_queue.clear()
        return steering

    def _dequeue_follow_up_messages(self) -> List[AgentMessage]:
        if self.follow_up_mode == "one-at-a-time":
            if self.follow_up_queue:
                return [self.follow_up_queue.pop(0)]
            return []
        follow_up = list(self.follow_up_queue)
        self.follow_up_queue.clear()
        return follow_up

    def clear_messages(self):
        self._state.messages.clear()

    def abort(self):
        if self.abort_controller and not self.abort_controller.is_set():
            self.abort_controller.set()

    async def wait_for_idle(self):
        if self.running_prompt and not self.running_prompt.done():
            await self.running_prompt

    def reset(self):
        self._state.messages.clear()
        self._state.is_streaming = False
        self._state.stream_message = None
        self._state.pending_tool_calls.clear()
        self._state.error = None
        self.clear_all_queues()

    async def prompt(self, input_val: Union[str, AgentMessage, List[AgentMessage]], images: Optional[List[ImageContent]] = None):
        """Send a prompt with an AgentMessage, text, or multiple messages."""
        if self._state.is_streaming:
            raise RuntimeError(
                "Agent is already processing a prompt. Use steer() or followUp() to queue messages, or wait for completion."
            )

        model = self._state.model
        if not model:
            raise ValueError("No model configured")

        msgs: List[AgentMessage]

        if isinstance(input_val, list):
            msgs = input_val
        elif isinstance(input_val, str):
            content: List[Union[TextContent, ImageContent]] = [{"type": "text", "text": input_val}] # type: ignore
            if images:
                content.extend(images)
            msgs = [{
                "role": "user",
                "content": content,
                "timestamp": int(time.time() * 1000)
            }] # type: ignore
        else:
            msgs = [input_val]

        await self._run_loop(msgs)

    async def continue_session(self):
        """Continue from current context (used for retries and resuming queued messages)."""
        if self._state.is_streaming:
            raise RuntimeError("Agent is already processing. Wait for completion before continuing.")

        messages = self._state.messages
        if not messages:
            raise ValueError("No messages to continue from")
            
        role = messages[-1].get("role") if isinstance(messages[-1], dict) else getattr(messages[-1], "role", None)
        if role == "assistant":
            queued_steering = self._dequeue_steering_messages()
            if queued_steering:
                await self._run_loop(queued_steering, skip_initial_steering_poll=True)
                return

            queued_follow_up = self._dequeue_follow_up_messages()
            if queued_follow_up:
                await self._run_loop(queued_follow_up)
                return

            raise ValueError("Cannot continue from message role: assistant")

        await self._run_loop(None)

    async def _run_loop(self, messages: Optional[List[AgentMessage]], skip_initial_steering_poll: bool = False):
        model = self._state.model
        if not model:
            raise ValueError("No model configured")

        self.abort_controller = asyncio.Event()
        self._state.is_streaming = True
        self._state.stream_message = None
        self._state.error = None

        reasoning = None if self._state.thinking_level == "off" else self._state.thinking_level

        context = AgentContext(
            system_prompt=self._state.system_prompt,
            messages=list(self._state.messages),
            tools=self._state.tools
        )

        async def get_steering_messages() -> List[AgentMessage]:
            nonlocal skip_initial_steering_poll
            if skip_initial_steering_poll:
                skip_initial_steering_poll = False
                return []
            return self._dequeue_steering_messages()

        async def get_follow_up_messages() -> List[AgentMessage]:
            return self._dequeue_follow_up_messages()

        effort = ReasoningEffort(reasoning) if reasoning and reasoning != "none" else ReasoningEffort.NONE

        config = AgentLoopConfig(
            model=model,
            reasoning_effort=effort,
            thinking_enabled=(effort != ReasoningEffort.NONE),
            thinking_budget=10_000 if effort != ReasoningEffort.NONE else 0,
            session_id=self._session_id,
            convert_to_llm=self.convert_to_llm,
            transform_context=self.transform_context,
            get_api_key=self.get_api_key,
            get_steering_messages=get_steering_messages,
            get_follow_up_messages=get_follow_up_messages,
            thinking_budgets=self._thinking_budgets
        )

        partial = None
        
        # We track running_prompt internally
        async def run():
            nonlocal partial
            try:
                stream = agent_loop(messages, context, config, self.stream_fn) if messages else agent_loop_continue(context, config, self.stream_fn)
                
                async for event in stream:
                    # Update internal state based on events
                    event_type = getattr(event, "type", None)
                    
                    if event_type == "message_start":
                        partial = event.message # type: ignore
                        self._state.stream_message = event.message # type: ignore
                    elif event_type == "message_update":
                        partial = event.message # type: ignore
                        self._state.stream_message = event.message # type: ignore
                    elif event_type == "message_end":
                        partial = None
                        self._state.stream_message = None
                        self.append_message(event.message) # type: ignore
                    elif event_type == "tool_execution_start":
                        self._state.pending_tool_calls.add(event.tool_call_id) # type: ignore
                    elif event_type == "tool_execution_end":
                        self._state.pending_tool_calls.discard(event.tool_call_id) # type: ignore
                    elif event_type == "turn_end":
                        if getattr(event.message, "role", None) == "assistant" and getattr(event.message, "error_message", None): # type: ignore
                            self._state.error = getattr(event.message, "error_message", None) # type: ignore
                    elif event_type == "agent_end":
                        self._state.is_streaming = False
                        self._state.stream_message = None
                    
                    self._emit(event)
                    
            except Exception as err:
                from sentarc_ai.types import AssistantMessage
                error_msg = AssistantMessage(
                    role="assistant",
                    content=[TextContent(type="text", text="")], # type: ignore
                    api=model.api, # type: ignore
                    provider=model.provider, # type: ignore
                    model=model.id, # type: ignore
                    usage={"input": 0, "output": 0, "cacheRead": 0, "cacheWrite": 0, "totalTokens": 0}, # type: ignore
                    stop_reason="aborted" if self.abort_controller and self.abort_controller.is_set() else "error",
                    error_message=str(err),
                    timestamp=int(time.time() * 1000)
                )
                self.append_message(error_msg) # type: ignore
                self._state.error = str(err)
                from sentarc_agent.types import AgentEndEvent
                self._emit(AgentEndEvent(messages=[error_msg])) # type: ignore
            finally:
                self._state.is_streaming = False
                self._state.stream_message = None
                self._state.pending_tool_calls.clear()
                self.abort_controller = None

        self.running_prompt = asyncio.create_task(run())
        await self.running_prompt

    def _emit(self, e: AgentEvent):
        for listener in self.listeners:
            listener(e)



import pytest
import asyncio

from sentarc_ai.types import (
    AssistantMessage,
    ModelDef,
    TextContent,
    ToolCallContent,
    ToolResultMessage
)
from sentarc_agent.types import (
    AgentEvent,
    AgentMessage,
    AgentState,
    AgentOptions,
    AgentLoopConfig,
    AgentTool,
    AgentToolResult
)
from sentarc_agent.agent import Agent, AgentOptions


# Mock model that returns straight text
class MockModel(ModelDef):
    def __init__(self, override_responses: list = None):
        super().__init__(id="mock", provider="test", api="test")
        self.responses = override_responses or [
            AssistantMessage(
                role="assistant",
                content=[TextContent(type="text", text="Hello back!")],
                api="test",
                provider="test",
                model="mock",
                usage={"input": 0, "output": 0, "cacheRead": 0, "cacheWrite": 0, "totalTokens": 0},
                stop_reason="stop"
            )
        ]
        self.call_count = 0

async def mock_stream_simple(model: ModelDef, context: dict, options=None, **kwargs):
    if isinstance(model, MockModel):
        res = model.responses[model.call_count]
        model.call_count += 1
        
        events = [
            type("Event", (), {"type": "start", "partial": res})(),
            type("Event", (), {"type": "text_start", "partial": res, "contentIndex": 0})(),
            type("Event", (), {"type": "text_delta", "partial": res, "contentIndex": 0, "delta": "Hello"})(),
            type("Event", (), {"type": "text_end", "partial": res, "contentIndex": 0})(),
            type("Event", (), {"type": "done", "message": res})()
        ]
        
        for e in events:
            yield e
        return
        
    raise ValueError("Not a mock model")


@pytest.mark.asyncio
async def test_agent_initialization():
    agent = Agent()
    assert agent.state.messages == []
    assert not agent.state.is_streaming
    assert agent.state.pending_tool_calls == set()

@pytest.mark.asyncio
async def test_simple_prompt():
    events = []
    
    agent = Agent(AgentOptions(
        stream_fn=mock_stream_simple
    ))
    agent.set_model(MockModel())
    
    agent.subscribe(lambda e: events.append(getattr(e, "type", type(e).__name__)))
    
    await agent.prompt("Hello")
    
    # Assert lifecycle events hit properly
    event_types = [e for e in events]
    assert "agent_start" in event_types
    assert "turn_start" in event_types
    assert "message_start" in event_types  # user prompt
    assert "message_update" in event_types
    assert "message_end" in event_types    # assistant answer
    assert "turn_end" in event_types
    assert "agent_end" in event_types
    
    # Check states
    assert len(agent.state.messages) == 2
    assert agent.state.messages[0]["role"] == "user" # type: ignore
    assert getattr(agent.state.messages[1], "role") == "assistant"
    
@pytest.mark.asyncio
async def test_tool_execution():
    async def mock_tool_execute(tool_call_id, args, signal, on_update):
        return AgentToolResult(
            content=[TextContent(type="text", text=f"Executed with {args['prop']}")],
            details={"prop": args["prop"]}
        )

    tool = AgentTool(
        name="test_tool",
        label="Test Tool",
        description="A test tool",
        parameters={"type": "object", "properties": {"prop": {"type": "string"}}},
        execute=mock_tool_execute
    )
    
    # Mock model will first return a tool call, then a text response
    tool_call_msg = AssistantMessage(
        role="assistant",
        content=[ToolCallContent(type="toolCall", id="call_1", name="test_tool", arguments={"prop": "test_val"})],
        api="test",
        provider="test",
        model="mock",
        usage={"input": 0, "output": 0, "cacheRead": 0, "cacheWrite": 0, "totalTokens": 0},
        stop_reason="toolUse"
    )
    final_text_msg = AssistantMessage(
        role="assistant",
        content=[TextContent(type="text", text="Done with tool")],
        api="test",
        provider="test",
        model="mock",
        usage={"input": 0, "output": 0, "cacheRead": 0, "cacheWrite": 0, "totalTokens": 0},
        stop_reason="stop"
    )
    
    agent = Agent(AgentOptions(stream_fn=mock_stream_simple))
    agent.set_model(MockModel([tool_call_msg, final_text_msg]))
    agent.set_tools([tool])
    
    events = []
    agent.subscribe(lambda e: events.append(getattr(e, "type", type(e).__name__)))
    
    await agent.prompt("Execute tool")
    
    # Agent should have queued up a user message -> assistant toolcall -> toolresult -> assistant text.
    assert "tool_execution_start" in events
    assert "tool_execution_end" in events
    
    messages = agent.state.messages
    assert len(messages) == 4
    
    def _role(m): return m.get("role") if isinstance(m, dict) else getattr(m, "role", None)

    assert _role(messages[0]) == "user"
    assert _role(messages[1]) == "assistant"
    assert _role(messages[2]) == "tool"
    assert _role(messages[3]) == "assistant"

    # Ensure tool execution details mapped into tool result
    assert messages[2].content[0].text == "Executed with test_val" # type: ignore



"""
Core loop logic for processing the conversational graphs and yielding asynchronous AgentEvent streams.
Handles low-level execution mapping of Tool calls, Steering queues, and Context updates.
"""

import asyncio
from typing import AsyncGenerator, List, Optional, Any, Callable, Awaitable

from sentarc_ai.stream import stream_simple
from sentarc_ai.types import (
    Context,
    ToolCallContent,
    ToolResultMessage,
    AssistantMessage
)
from sentarc_ai.utils.validation import validate_tool_arguments

from sentarc_agent.types import (
    AgentContext,
    AgentLoopConfig,
    AgentMessage,
    AgentEvent,
    AgentStartEvent,
    AgentEndEvent,
    TurnStartEvent,
    TurnEndEvent,
    MessageStartEvent,
    MessageUpdateEvent,
    MessageEndEvent,
    ToolExecutionStartEvent,
    ToolExecutionUpdateEvent,
    ToolExecutionEndEvent,
    AgentToolResult,
    AgentTool,
    StreamFn
)

async def agent_loop(
    prompts: List[AgentMessage],
    context: AgentContext,
    config: AgentLoopConfig,
    stream_fn: Optional[StreamFn] = None
) -> AsyncGenerator[AgentEvent, None]:
    """
    Start an agent loop with a new prompt message.
    The prompt is added to the context and events are emitted for it.
    """
    new_messages: List[AgentMessage] = list(prompts)
    current_context = AgentContext(
        system_prompt=context.system_prompt,
        messages=context.messages + prompts,
        tools=context.tools
    )

    yield AgentStartEvent()
    yield TurnStartEvent()
    
    for prompt in prompts:
        yield MessageStartEvent(message=prompt)
        yield MessageEndEvent(message=prompt)

    async for event in _run_loop(current_context, new_messages, config, stream_fn):
        yield event


async def agent_loop_continue(
    context: AgentContext,
    config: AgentLoopConfig,
    stream_fn: Optional[StreamFn] = None
) -> AsyncGenerator[AgentEvent, None]:
    """
    Continue an agent loop from the current context without adding a new message.
    Used for retries - context already has user message or tool results.
    """
    if not context.messages:
        raise ValueError("Cannot continue: no messages in context")
    
    last_message = context.messages[-1]
    # Simple check for basic dictionary structure or role attribute
    role = last_message.get("role") if isinstance(last_message, dict) else getattr(last_message, "role", None)
    
    if role == "assistant":
        raise ValueError("Cannot continue from message role: assistant")

    new_messages: List[AgentMessage] = []
    current_context = AgentContext(
        system_prompt=context.system_prompt,
        messages=list(context.messages),
        tools=context.tools
    )

    yield AgentStartEvent()
    yield TurnStartEvent()

    async for event in _run_loop(current_context, new_messages, config, stream_fn):
        yield event


async def _run_loop(
    current_context: AgentContext,
    new_messages: List[AgentMessage],
    config: AgentLoopConfig,
    stream_fn: Optional[StreamFn] = None
) -> AsyncGenerator[AgentEvent, None]:
    """Main loop logic shared by agent_loop and agent_loop_continue."""
    first_turn = True
    
    pending_messages: List[AgentMessage] = []
    if config.get_steering_messages:
        pending_messages = await config.get_steering_messages()

    while True:
        has_more_tool_calls = True
        steering_after_tools: Optional[List[AgentMessage]] = None

        while has_more_tool_calls or pending_messages:
            if not first_turn:
                yield TurnStartEvent()
            else:
                first_turn = False

            # Process pending messages (inject before next assistant response)
            if pending_messages:
                for message in pending_messages:
                    yield MessageStartEvent(message=message) # type: ignore
                    yield MessageEndEvent(message=message) # type: ignore
                    current_context.messages.append(message)
                    new_messages.append(message)
                pending_messages = []

            # Stream assistant response
            message = None
            async for event in _stream_assistant_response(current_context, config, stream_fn):
                if isinstance(event, AssistantMessage):
                    message = event
                else:
                    yield event
                    
            if not message:
                raise RuntimeError("Failed to generate assistant message.")
                
            new_messages.append(message)

            if getattr(message, "stop_reason", None) in ["error", "aborted"]:
                yield TurnEndEvent(message=message, tool_results=[])
                yield AgentEndEvent(messages=new_messages)
                return

            # Check for tool calls
            tool_calls = [c for c in message.content if isinstance(c, ToolCallContent)]
            has_more_tool_calls = len(tool_calls) > 0

            tool_results: List[ToolResultMessage] = []
            if has_more_tool_calls:
                async for event in _execute_tool_calls(
                    current_context.tools, message, config.get_steering_messages
                ):
                    if isinstance(event, tuple):
                        tool_results.extend(event[0])
                        steering_after_tools = event[1]
                    else:
                        yield event

                for result in tool_results:
                    current_context.messages.append(result)
                    new_messages.append(result)

            yield TurnEndEvent(message=message, tool_results=tool_results)

            # Get steering messages after turn completes
            if steering_after_tools and len(steering_after_tools) > 0:
                pending_messages = steering_after_tools
                steering_after_tools = None
            elif config.get_steering_messages:
                pending_messages = await config.get_steering_messages()

        # Check for follow-up messages
        follow_up_messages: List[AgentMessage] = []
        if config.get_follow_up_messages:
            follow_up_messages = await config.get_follow_up_messages()
            
        if follow_up_messages:
            pending_messages = follow_up_messages
            continue

        break

    yield AgentEndEvent(messages=new_messages)


async def _stream_assistant_response(
    context: AgentContext,
    config: AgentLoopConfig,
    stream_fn: Optional[StreamFn] = None
) -> AsyncGenerator[AgentEvent | AssistantMessage, None]:
    """Streams an assistant response, yielding AgentEvents and finally the AssistantMessage."""
    messages = context.messages
    if config.transform_context:
        messages = await config.transform_context(messages, None)

    if not config.convert_to_llm:
        raise ValueError("convert_to_llm function is required in AgentLoopConfig")
        
    llm_messages = config.convert_to_llm(messages)
    if asyncio.iscoroutine(llm_messages):
        llm_messages = await llm_messages # type: ignore

    llm_context = Context(
        system_prompt=context.system_prompt,
        messages=llm_messages, # type: ignore
        tools=context.tools # type: ignore
    )

    stream_function = stream_fn or stream_simple

    resolved_api_key = config.api_key
    if config.get_api_key:
        api_key_result = config.get_api_key(config.model.provider)
        if asyncio.iscoroutine(api_key_result):
            resolved_api_key = await api_key_result # type: ignore
        elif api_key_result:
             resolved_api_key = api_key_result # type: ignore

    config.api_key = resolved_api_key

    # Pass the event stream async generator correctly
    response_stream = stream_function(config.model, llm_context, config)

    partial_message = None
    added_partial = False

    async for event in response_stream:
        if event.type == "start":
            partial_message = event.partial
            context.messages.append(partial_message)
            added_partial = True
            yield MessageStartEvent(message=partial_message) # type: ignore
            
        elif event.type in ["text_start", "text_delta", "text_end", 
                            "thinking_start", "thinking_delta", "thinking_end", 
                            "toolcall_start", "toolcall_delta", "toolcall_end"]:
            if partial_message:
                partial_message = event.partial
                context.messages[-1] = partial_message
                yield MessageUpdateEvent(
                    message=partial_message, # type: ignore
                    assistant_message_event=event
                )
                
        elif event.type in ["done", "error"]:
            # Depending on python implementation, `error` event might just emit the finalized error AssitantMessage
            final_message = event.message if event.type == "done" else event.error
            if added_partial:
                context.messages[-1] = final_message
            else:
                context.messages.append(final_message)
                yield MessageStartEvent(message=final_message) # type: ignore
            
            yield MessageEndEvent(message=final_message) # type: ignore
            yield final_message # Explicitly return the finalized AssistantMessage
            return


async def _execute_tool_calls(
    tools: Optional[List[AgentTool]],
    assistant_message: AssistantMessage,
    get_steering_messages: Optional[Callable[[], Awaitable[List[AgentMessage]]]]
) -> AsyncGenerator[Any, None]: # Yields AgentEvents, then finally (tool_results, steering_messages)
    tool_calls = [c for c in assistant_message.content if isinstance(c, ToolCallContent)]
    results: List[ToolResultMessage] = []
    steering_messages: Optional[List[AgentMessage]] = None

    for index, tool_call in enumerate(tool_calls):
        tool = next((t for t in (tools or []) if t.name == tool_call.name), None)

        yield ToolExecutionStartEvent(
            tool_call_id=tool_call.id,
            tool_name=tool_call.name,
            args=tool_call.arguments
        )

        result: AgentToolResult
        is_error = False

        try:
            if not tool:
                raise ValueError(f"Tool {tool_call.name} not found")

            # Validate arguments mapping dictionaries natively using json schema utility port
            validated_args = validate_tool_arguments(tool, tool_call)

            def on_update(partial_result: AgentToolResult):
                return ToolExecutionUpdateEvent(
                    tool_call_id=tool_call.id,
                    tool_name=tool_call.name,
                    args=tool_call.arguments,
                    partial_result=partial_result
                )
            # In purely async python, bridging sync callbacks to event yielders can be tricky.
            # We skip intermediate partial yielding inside the execute function here for simplicity 
            # unless the tool specifically queues it.

            exec_res = tool.execute(tool_call.id, validated_args, None, None)
            if asyncio.iscoroutine(exec_res):
                result = await exec_res
            else:
                result = exec_res # type: ignore

        except Exception as e:
            result = AgentToolResult(
                content=[TextContent(type="text", text=str(e))],
                details={}
            )
            is_error = True

        yield ToolExecutionEndEvent(
            tool_call_id=tool_call.id,
            tool_name=tool_call.name,
            result=result,
            is_error=is_error
        )

        import time
        tool_result_message = ToolResultMessage(
            role="tool",
            tool_call_id=tool_call.id,
            content=result.content,
            is_error=is_error,
            timestamp=int(time.time() * 1000)
        )

        results.append(tool_result_message)
        yield MessageStartEvent(message=tool_result_message) # type: ignore
        yield MessageEndEvent(message=tool_result_message) # type: ignore

        # Check for steering mid-tool executions
        if get_steering_messages:
            steering = await get_steering_messages()
            if steering:
                steering_messages = steering
                remaining_calls = tool_calls[index + 1:]
                for skipped in remaining_calls:
                    skipped_res = _skip_tool_call(skipped)
                    for event in skipped_res[:-1]:
                        yield event
                    results.append(skipped_res[-1]) # type: ignore
                break

    yield (results, steering_messages)


def _skip_tool_call(tool_call: ToolCallContent) -> tuple:
    result = AgentToolResult(
        content=[TextContent(type="text", text="Skipped due to queued user message.")],
        details={}
    )
    
    events = [
        ToolExecutionStartEvent(
            tool_call_id=tool_call.id,
            tool_name=tool_call.name,
            args=tool_call.arguments
        ),
        ToolExecutionEndEvent(
            tool_call_id=tool_call.id,
            tool_name=tool_call.name,
            result=result,
            is_error=True
        )
    ]
    
    import time
    tool_result_message = ToolResultMessage(
        role="tool",
        tool_call_id=tool_call.id,
        content=result.content,
        is_error=True,
        timestamp=int(time.time() * 1000)
    )
    
    events.extend([
        MessageStartEvent(message=tool_result_message), # type: ignore
        MessageEndEvent(message=tool_result_message) # type: ignore
    ])
    
    events.append(tool_result_message)
    return tuple(events)



"""
Core types, dataclasses, and event discriminators for the sentarc-agent package.
Provides type safety for the overall agent loop, streaming events, and tool configurations.
"""

from typing import Any, Callable, Dict, List, Literal, Optional, Set, Union, Awaitable
from dataclasses import dataclass, field
import asyncio

from sentarc_ai.types import (
    AssistantMessage,
    StreamEvent,
    Context,
    ImageContent,
    Message,
    ModelDef,
    StreamOptions,
    TextContent,
    Tool,
    ToolCallContent,
    ToolResultMessage,
    ReasoningEffort,
)

# In Python, we can't easily do declaration merging for CustomAgentMessages like in TS.
# We'll rely on `Dict[str, Any]` or a base class `CustomAgentMessage` for users to extend.
# For simplicity and parity with the TS union `Message | CustomAgentMessages[keyof CustomAgentMessages]`,
# we define `AgentMessage` as a union of standard `Message` and `Any` (allowing custom dicts/dataclasses).
AgentMessage = Union[Message, Dict[str, Any]]

ThinkingLevel = Literal["off", "minimal", "low", "medium", "high", "xhigh"]


@dataclass
class AgentToolResult:
    """Result of an agent tool execution."""
    content: List[Union[TextContent, ImageContent]]
    details: Any


# Callback for streaming tool execution updates
AgentToolUpdateCallback = Callable[[AgentToolResult], None]


@dataclass
class AgentTool(Tool):
    """
    Extends sentarc_ai.types.Tool with an execute function and a label.
    """
    label: str = ""
    
    # execute signature: (tool_call_id: str, args: dict, signal: asyncio.Event, on_update: callable) -> AgentToolResult
    # Using Awaitable here to signify it's an async function
    execute: Callable[[str, Dict[str, Any], Optional[asyncio.Event], Optional[AgentToolUpdateCallback]], Awaitable[AgentToolResult]] = None # type: ignore


@dataclass
class AgentContext:
    system_prompt: str
    messages: List[AgentMessage]
    tools: Optional[List[AgentTool]] = None


@dataclass
class AgentState:
    system_prompt: str
    model: Optional[ModelDef]
    thinking_level: ThinkingLevel
    tools: List[AgentTool]
    messages: List[AgentMessage]
    is_streaming: bool
    stream_message: Optional[AgentMessage]
    pending_tool_calls: Set[str]
    error: Optional[str] = None


@dataclass
class AgentLoopConfig(StreamOptions):
    model: ModelDef = None # type: ignore
    thinking_budgets: Optional[Dict[str, int]] = None
    
    # Converts AgentMessage[] to LLM-compatible Message[] before each LLM call.
    convert_to_llm: Callable[[List[AgentMessage]], Union[List[Message], Awaitable[List[Message]]]] = None # type: ignore

    # Optional transform applied to the context before `convertToLlm`.
    transform_context: Optional[Callable[[List[AgentMessage], Optional[asyncio.Event]], Awaitable[List[AgentMessage]]]] = None

    # Resolves an API key dynamically for each LLM call.
    get_api_key: Optional[Callable[[str], Union[Optional[str], Awaitable[Optional[str]]]]] = None

    # Returns steering messages to inject into the conversation mid-run.
    get_steering_messages: Optional[Callable[[], Awaitable[List[AgentMessage]]]] = None

    # Returns follow-up messages to process after the agent would otherwise stop.
    get_follow_up_messages: Optional[Callable[[], Awaitable[List[AgentMessage]]]] = None


# Events emitted by the Agent for UI updates.

@dataclass
class AgentStartEvent:
    type: Literal["agent_start"] = "agent_start"

@dataclass
class AgentEndEvent:
    messages: List[AgentMessage]
    type: Literal["agent_end"] = "agent_end"

@dataclass
class TurnStartEvent:
    type: Literal["turn_start"] = "turn_start"

@dataclass
class TurnEndEvent:
    message: AgentMessage
    tool_results: List[ToolResultMessage]
    type: Literal["turn_end"] = "turn_end"

@dataclass
class MessageStartEvent:
    message: AgentMessage
    type: Literal["message_start"] = "message_start"

@dataclass
class MessageUpdateEvent:
    message: AgentMessage
    assistant_message_event: StreamEvent
    type: Literal["message_update"] = "message_update"

@dataclass
class MessageEndEvent:
    message: AgentMessage
    type: Literal["message_end"] = "message_end"

@dataclass
class ToolExecutionStartEvent:
    tool_call_id: str
    tool_name: str
    args: Any
    type: Literal["tool_execution_start"] = "tool_execution_start"

@dataclass
class ToolExecutionUpdateEvent:
    tool_call_id: str
    tool_name: str
    args: Any
    partial_result: Any
    type: Literal["tool_execution_update"] = "tool_execution_update"

@dataclass
class ToolExecutionEndEvent:
    tool_call_id: str
    tool_name: str
    result: Any
    is_error: bool
    type: Literal["tool_execution_end"] = "tool_execution_end"


AgentEvent = Union[
    AgentStartEvent,
    AgentEndEvent,
    TurnStartEvent,
    TurnEndEvent,
    MessageStartEvent,
    MessageUpdateEvent,
    MessageEndEvent,
    ToolExecutionStartEvent,
    ToolExecutionUpdateEvent,
    ToolExecutionEndEvent,
]

# A signature for the stream_fn. It aligns with stream_simple from sentarc_ai.stream
StreamFn = Callable[..., Any] # We'll keep this loose in Python to avoid circular deps if needed, passing stream_simple dynamically usually.

@dataclass
class AgentOptions:
    initial_state: Optional[Dict[str, Any]] = None
    convert_to_llm: Optional[Callable[[List[AgentMessage]], Union[List[Message], Awaitable[List[Message]]]]] = None
    transform_context: Optional[Callable[[List[AgentMessage], Optional[asyncio.Event]], Awaitable[List[AgentMessage]]]] = None
    steering_mode: Optional[Literal["all", "one-at-a-time"]] = None
    follow_up_mode: Optional[Literal["all", "one-at-a-time"]] = None
    stream_fn: Optional[StreamFn] = None
    session_id: Optional[str] = None
    get_api_key: Optional[Callable[[str], Union[Optional[str], Awaitable[Optional[str]]]]] = None
    thinking_budgets: Optional[Dict[str, int]] = None
